// © 2023 John Breauxs
//! Common error type for [msp430-asm](crate) errors

use std::fmt::Display;

use super::{
    tokenizer::token::{OwnedToken, Types},
    *,
};

// TODO: Store more error context in error. for example:
// Error {ExpectationFailed{...}, WhileParsing(Register)}

#[derive(Debug)]
pub enum Error {
    /// Produced by [Parser](crate::parser::Parser::parse<T>())
    ParseError(parser::root::Root, Box<dyn std::error::Error + 'static>),
    Contextual(Context, Box<Self>),
    /// Produced by [Token] when the input is entirely unexpected.
    UnexpectedSymbol(String),
    /// Produced by [`TokenStream::expect`] when the next [Token] isn't the expected [Type]
    UnexpectedToken {
        expected: Type,
        got: OwnedToken,
    },
    /// Produced by [`TokenStream::expect_any_of`] when the next [Token] isn't any of the expected
    /// [Types](Type)
    AllExpectationsFailed {
        expected: Types,
        got: OwnedToken,
    },
    /// Produced by
    /// [Number](parser::instruction::encoding::number::Number)[::parse()](parser::parsable::Parsable::parse())
    /// when the parsed number contains digits too high for the specified radix
    UnexpectedDigits(String, u32),
    /// Produced by
    /// [Opcode](parser::instruction::opcode::Opcode)[::parse()](parser::parsable::Parsable::parse())
    /// when the opcode passed lexing but did not match recognized opcodes.
    ///
    /// This should be interpreted as a failure in lexing.
    UnrecognizedOpcode(String),
    /// Produced by [Register](parser::instruction::encoding::register::Register)
    /// when attempting to convert from a [str] that isn't a register (pc, sp, sr, cg, or r{number})
    NotARegister(String),
    /// Produced by [Register](parser::instruction::encoding::register)
    /// when attempting to convert from a [u16] that isn't in the range 0-15
    RegisterTooHigh(u16),
    /// Produced by
    /// [SecondaryOperand](parser::instruction::encoding::secondary_operand)
    /// when the joke "secondary immediate" form is out of range 0..=1
    FatSecondaryImmediate(isize),
    /// Produced by [Number](parser::instruction::encoding::number) when the number is too
    /// wide to fit in 16 bits (outside the range `(-2^15) .. (2^16-1)` )
    NumberTooWide(isize),
    /// Produced by [JumpTarget](parser::instruction::encoding::jump_target)
    /// when the jump offset is outside the range (-0x3ff..0x3fc)
    JumpedTooFar(isize),
    /// Produced by [JumpTarget](parser::instruction::encoding::jump_target)
    JumpedOdd(isize),
    EndOfFile,
}

impl Error {
    pub fn context(self, c: Context) -> Self {
        match self {
            Self::Contextual(..) => self,
            _ => Self::Contextual(c, Box::new(self)),
        }
    }

    // Extracts the root of the error tree
    pub fn bare(self) -> Self {
        match self {
            Self::Contextual(_, bare) => bare.bare(),
            _ => self,
        }
    }

    pub fn swap(mut self, other: Self) -> Self {
        if let Self::Contextual(_, err) = &mut self {
            _ = std::mem::replace(err.as_mut(), other)
        }
        self
    }

    pub fn expected<E: AsRef<[Type]>, T: Into<OwnedToken>>(expected: E, got: T) -> Self {
        match expected.as_ref().len() {
            1 => Self::UnexpectedToken { expected: expected.as_ref()[0], got: got.into() },
            _ => Self::AllExpectationsFailed { expected: expected.as_ref().into(), got: got.into() },
        }
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Error::Contextual(ctx, error) => write!(f, "{ctx}: {error}"),
            Error::ParseError(_, error) => write!(f, "{error}"),
            Error::UnexpectedSymbol(sym) => write!(f, "Unexpected item in bagging area: \"{sym}\""),
            Error::UnexpectedToken { expected, got } => write!(f, "Expected {expected}, got {got}."),
            Error::AllExpectationsFailed { expected, got } => write!(f, "Expected {expected}, got {got}."),
            Error::UnexpectedDigits(number, radix) => write!(f, "Number `{number}` is not base {radix}."),
            Error::UnrecognizedOpcode(op) => write!(f, "{op} is not an opcode"),
            Error::NotARegister(reg) => write!(f, "{reg} is not a register"),
            Error::RegisterTooHigh(reg) => write!(f, "r{reg} is not a register"),
            Error::FatSecondaryImmediate(num) => write!(f, "Secondary immediate must be #0 or #1, not #{num}"),
            Error::NumberTooWide(num) => write!(f, "{num} does not fit in 16 bits"),
            Error::JumpedTooFar(num) => write!(f, "{num} is too far away: must be in range (`-1022..=1024`.)"),
            Error::JumpedOdd(num) => write!(f, "Jump targets only encode even numbers: {num} must not be odd."),
            Error::EndOfFile => write!(f, "Unexpected end of file"),
        }
    }
}

impl std::error::Error for Error {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            Error::ParseError(_, e) => Some(e.as_ref()),
            _ => None,
        }
    }
}
