// © 2023 John Breaux
//! An [Instruction] contains the [Opcode] and [Encoding] information for a single msp430
//! instruction
//!
//!
//! Note: [Opcode] and [Encoding] are very tightly coupled, because they represent interdependent
//! parts of the same instruction. This is why [Opcode]::resolve() returns an [EncodingParser] --
//! otherwise, there's an explosion of states that I can't really cope with on my own. Really,
//! there's about 9 valid classes of instruction, some of which are only used for one or two of the
//! MSP430's instructions.

use super::*;

pub mod encoding;
pub mod opcode;

/// Represents an entire MSP430 instruction
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct Instruction(Opcode, Encoding);

impl Instruction {
    pub fn opcode(&self) -> &Opcode { &self.0 }
    pub fn encoding(&self) -> &Encoding { &self.1 }
    /// Gets the Instruction as a [u16]
    pub fn word(&self) -> u16 { self.0 as u16 | self.1.word() }
    /// Gets the [extension words]
    pub fn ext_words(&self) -> (Option<u16>, Option<u16>) { self.1.extwords() }
}

impl Parsable for Instruction {
    fn parse<'text, T>(p: &Parser, stream: &mut T) -> Result<Self, Error>
    where
        Self: Sized,
        T: crate::TokenStream<'text>,
    {
        // parse an opcode
        let opcode: Opcode = Opcode::parse(p, stream)?;
        // resolve the opcode to a final opcode and an encoding
        let (opcode, encoding) = opcode.resolve();
        // parse the encoding
        let encoding = encoding.parse(p, stream)?;
        Ok(Self(opcode, encoding))
    }
}

impl From<Instruction> for u16 {
    fn from(value: Instruction) -> Self { value.word() }
}

impl Display for Instruction {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result { write!(f, "{}{}", self.0, self.1) }
}

impl LowerHex for Instruction {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let (word, (ext_src, ext_dst)) = (self.word(), self.ext_words());
        write!(f, "{:04x} ", word.swap_bytes())?;
        if let Some(e) = ext_src {
            write!(f, "{:04x} ", e.swap_bytes())?
        }
        if let Some(e) = ext_dst {
            write!(f, "{:04x} ", e.swap_bytes())?
        }
        Ok(())
    }
}
