// © 2023 John Breaux
//! A [Number] represents a 16-bit signed or unsigned word
use super::*;

// TODO: Allow identifiers/expressions  in place of numbers
//       - Dependency inversion in TokenStream to allow swapping the parser mid-parse?
//       - Oh my god, not relying on std::iter::Iterator allows for so many more parsing options

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct Number(isize, u32); // (value, radix)

impl Parsable for Number {
    // A number is:
    // RadixMarker[Hex|Oct|Bin]?
    // - Number
    fn parse<'text, T>(p: &Parser, stream: &mut T) -> Result<Self, Error>
    where T: TokenStream<'text> {
        use Type::*;
        // The number is negative when it begins with a Minus, but Plus is also acceptable.
        let negative = stream.expect_any_of([Minus, Plus]).map_or(false, |t| t.is_variant(Minus));
        let radix = match stream
            .expect_any_of([RadixMarkerHex, RadixMarkerDec, RadixMarkerOct, RadixMarkerBin])
            .ok()
            .map(|t| t.variant())
        {
            Some(RadixMarkerHex) => 16,
            Some(RadixMarkerDec) => 10,
            Some(RadixMarkerOct) => 8,
            Some(RadixMarkerBin) => 2,
            _ => p.radix,
        };
        let number = stream.expect(Number)?;
        let number = isize::from_str_radix(number.lexeme(), radix)
            .map_err(|_| Error::UnexpectedDigits(number.lexeme().into(), radix).context(stream.context()))?
            * if negative { -1 } else { 1 };
        // Ensure number fits within a *signed or unsigned* 16-bit int (it will be truncated to fit)
        Ok(Self(
            if (-0x8000..0x10000).contains(&number) {
                number
            } else {
                Err(Error::NumberTooWide(number).context(stream.context()))?
            },
            radix,
        ))
    }
}
impl From<Number> for isize {
    fn from(value: Number) -> Self { value.0 as Self }
}
impl From<Number> for i32 {
    fn from(value: Number) -> Self { value.0 as Self }
}
impl From<Number> for u16 {
    /// Converts this type from the input type.
    fn from(value: Number) -> Self { value.0 as Self }
}

impl std::ops::Sub<isize> for Number {
    type Output = Self;
    fn sub(mut self, rhs: isize) -> Self::Output {
        self.0 -= rhs;
        self
    }
}

impl std::ops::Shr<usize> for Number {
    type Output = Self;
    fn shr(mut self, rhs: usize) -> Self::Output {
        self.0 >>= rhs;
        self
    }
}

impl std::fmt::Display for Number {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result { write!(f, "{:x}", self.0) }
}
