// © 2023 John Breaux
//! A [Width] represents whether an instruction operates on whole words or bytes
use super::*;

/// Represents an instruction's operand width.
///
/// Evaluates to false when instruction takes word-sized operands, or true when
/// instruction takes byte-sized operands
#[derive(Clone, Copy, Default, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct Width(bool);

impl Parsable for Width {
    fn parse<'text, T>(_: &Parser, stream: &mut T) -> Result<Self, Error>
    where T: TokenStream<'text> {
        let Ok(token) = stream.expect_any_of([Type::ByteWidth, Type::WordWidth]) else {
            return Ok(Self(false));
        };
        Ok(Self(token.is_variant(Type::ByteWidth)))
    }
}
impl From<Width> for u16 {
    fn from(value: Width) -> Self { (value.0 as Self) << 6 }
}
impl From<Width> for bool {
    fn from(value: Width) -> Self { value.0 }
}
impl From<bool> for Width {
    fn from(value: bool) -> Self { Width(value) }
}
impl std::fmt::Display for Width {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result { f.write_str(if self.0 { ".b" } else { "" }) }
}
