// © 2023 John Breaux
//! An [Opcode] encodes an msp430 operation
use super::*;

use std::str::FromStr;

/// Opcode from the [MSPGCC Manual][1]
///
/// Calling [`resolve()`](Opcode::resolve()) will emit an [EncodingParser] which will
/// extract from a [TokenStream] only the required arguments for that call.
///
/// [1]: https://mspgcc.sourceforge.net/manual/x223.html
#[allow(clippy::identity_op)]
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Opcode {
    // "Emulated" opcodes
    Nop,
    Pop,
    Br,
    Ret,
    Clrc,
    Setc,
    Clrz,
    Setz,
    Clrn,
    Setn,
    Dint,
    Eint,
    Rla,
    Rlc,
    Inv,
    Clr,
    Tst,
    Dec,
    Decd,
    Inc,
    Incd,
    Adc,
    Dadc,
    Sbc,
    // Single
    Rrc  = 0x1000 | 0 << 7,
    Swpb = 0x1000 | 1 << 7,
    Rra  = 0x1000 | 2 << 7,
    Sxt  = 0x1000 | 3 << 7,
    Push = 0x1000 | 4 << 7,
    Call = 0x1000 | 5 << 7,
    Reti = 0x1000 | 6 << 7,
    // Jump
    Jnz  = 0x2000 | 0 << 10,
    Jz   = 0x2000 | 1 << 10,
    Jnc  = 0x2000 | 2 << 10,
    Jc   = 0x2000 | 3 << 10,
    Jn   = 0x2000 | 4 << 10,
    Jge  = 0x2000 | 5 << 10,
    Jl   = 0x2000 | 6 << 10,
    Jmp  = 0x2000 | 7 << 10,
    // Double
    Mov  = 0x4000,
    Add  = 0x5000,
    Addc = 0x6000,
    Subc = 0x7000,
    Sub  = 0x8000,
    Cmp  = 0x9000,
    Dadd = 0xa000,
    Bit  = 0xb000,
    Bic  = 0xc000,
    Bis  = 0xd000,
    Xor  = 0xe000,
    And  = 0xf000,
}

impl Opcode {
    pub fn takes_width(&self) -> bool {
        use Opcode::*;
        match self {
            Rrc => true,
            Swpb => false,
            Rra => true,
            Sxt => false,
            Push => true,
            Call | Reti => false,
            Jnz | Jz | Jnc | Jc | Jn | Jge | Jl | Jmp => false,
            Mov | Add | Addc | Subc | Sub | Cmp | Dadd | Bit | Bic | Bis | Xor | And => true,
            Nop | Pop | Br | Ret | Clrc | Setc | Clrz | Setz | Clrn | Setn | Dint | Eint | Rla | Rlc | Inv | Clr
            | Tst | Dec | Decd | Inc | Incd | Adc | Dadc | Sbc => true,
        }
    }
    /// Resolve an Opcode into an [Opcode] and an [EncodingParser]
    pub fn resolve(self) -> (Opcode, EncodingParser) {
        use super::Encoding as Enc;
        use Opcode::*;
        use Register::*;
        use {PrimaryOperand as Src, SecondaryOperand as Dst};
        match self {
            Rrc | Swpb | Rra | Sxt | Push | Call | Reti => (self, Enc::single().end()),
            Jnz | Jz | Jnc | Jc | Jn | Jge | Jl | Jmp => (self, Enc::jump().end()),
            Mov | Add | Addc | Subc | Sub | Cmp | Dadd | Bit | Bic | Bis | Xor | And => (self, Enc::double().end()),
            Nop => (Mov, Enc::double().src(Src::Zero).dst(Dst::Zero).end()),
            Pop => (Mov, Enc::double().src(Src::PostInc(sp)).end()),
            Br => (Mov, Enc::double().dst(Dst::Direct(pc)).end()),
            Ret => (Mov, Enc::double().src(Src::PostInc(sp)).dst(Dst::Direct(pc)).end()),
            Clrc => (Bic, Enc::double().src(Src::One).dst(Dst::Direct(sr)).end()),
            Setc => (Bis, Enc::double().src(Src::One).dst(Dst::Direct(sr)).end()),
            Clrz => (Bic, Enc::double().src(Src::Two).dst(Dst::Direct(sr)).end()),
            Setz => (Bis, Enc::double().src(Src::Two).dst(Dst::Direct(sr)).end()),
            Clrn => (Bic, Enc::double().src(Src::Four).dst(Dst::Direct(sr)).end()),
            Setn => (Bis, Enc::double().src(Src::Four).dst(Dst::Direct(sr)).end()),
            Dint => (Bic, Enc::double().src(Src::Eight).dst(Dst::Direct(sr)).end()),
            Eint => (Bis, Enc::double().src(Src::Eight).dst(Dst::Direct(sr)).end()),
            Rla => (Add, Enc::reflexive().end()),
            Rlc => (Addc, Enc::reflexive().end()),
            Inv => (Xor, Enc::double().src(Src::MinusOne).end()),
            Clr => (Mov, Enc::double().src(Src::Zero).end()),
            Tst => (Cmp, Enc::double().src(Src::Zero).end()),
            Dec => (Sub, Enc::double().src(Src::One).end()),
            Decd => (Sub, Enc::double().src(Src::Two).end()),
            Inc => (Add, Enc::double().src(Src::One).end()),
            Incd => (Add, Enc::double().src(Src::Two).end()),
            Adc => (Addc, Enc::double().src(Src::Zero).end()),
            Dadc => (Dadd, Enc::double().src(Src::Zero).end()),
            Sbc => (Subc, Enc::double().src(Src::Zero).end()),
        }
    }
}

impl Parsable for Opcode {
    fn parse<'text, T>(_: &Parser, stream: &mut T) -> Result<Self, Error>
    where T: TokenStream<'text> {
        stream.expect(Type::Insn)?.parse().map_err(|e: Error| e.context(stream.context()))
    }
}

impl FromStr for Opcode {
    type Err = Error;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        use Opcode::*;
        //TODO: Reduce allocations here?
        let s = s.to_ascii_lowercase();
        Ok(match s.as_str() {
            "rrc" => Rrc,
            "swpb" => Swpb,
            "rra" => Rra,
            "sxt" => Sxt,
            "push" => Push,
            "call" => Call,
            "reti" => Reti,

            "jne" | "jnz" => Jnz,
            "jeq" | "jz" => Jz,
            "jnc" | "jlo" => Jnc,
            "jc" | "jhs" => Jc,
            "jn" => Jn,
            "jge" => Jge,
            "jl" => Jl,
            "jmp" => Jmp,

            "mov" => Mov,
            "add" => Add,
            "addc" => Addc,
            "subc" => Subc,
            "sub" => Sub,
            "cmp" => Cmp,
            "dadd" => Dadd,
            "bit" => Bit,
            "bic" => Bic,
            "bis" => Bis,
            "xor" => Xor,
            "and" => And,

            "nop" => Nop,
            "pop" => Pop,
            "br" => Br,
            "ret" => Ret,
            "clrc" => Clrc,
            "setc" => Setc,
            "clrz" => Clrz,
            "setz" => Setz,
            "clrn" => Clrn,
            "setn" => Setn,
            "dint" => Dint,
            "eint" => Eint,
            "rla" => Rla,
            "rlc" => Rlc,
            "inv" => Inv,
            "clr" => Clr,
            "tst" => Tst,
            "dec" => Dec,
            "decd" => Decd,
            "inc" => Inc,
            "incd" => Incd,
            "adc" => Adc,
            "dadc" => Dadc,
            "sbc" => Sbc,
            _ => Err(Error::UnrecognizedOpcode(s))?,
        })
    }
}

impl Display for Opcode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        use Opcode::*;
        write!(
            f,
            "{}",
            match self {
                Nop => "nop",
                Pop => "pop",
                Br => "br",
                Ret => "ret",
                Clrc => "clrc",
                Setc => "setc",
                Clrz => "clrz",
                Setz => "setz",
                Clrn => "clrn",
                Setn => "setn",
                Dint => "dint",
                Eint => "eint",
                Rla => "rla",
                Rlc => "rlc",
                Inv => "inv",
                Clr => "clr",
                Tst => "tst",
                Dec => "dec",
                Decd => "decd",
                Inc => "inc",
                Incd => "incd",
                Adc => "adc",
                Dadc => "dadc",
                Sbc => "sbc",
                Rrc => "rrc",
                Swpb => "swpb",
                Rra => "rra",
                Sxt => "sxt",
                Push => "push",
                Call => "call",
                Reti => "reti",
                Jnz => "jnz",
                Jz => "jz",
                Jnc => "jnc",
                Jc => "jc",
                Jn => "jn",
                Jge => "jge",
                Jl => "jl",
                Jmp => "jmp",
                Mov => "mov",
                Add => "add",
                Addc => "addc",
                Subc => "subc",
                Sub => "sub",
                Cmp => "cmp",
                Dadd => "dadd",
                Bit => "bit",
                Bic => "bic",
                Bis => "bis",
                Xor => "xor",
                And => "and",
            }
        )
    }
}

impl LowerHex for Opcode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result { write!(f, "{:04x}", *self as u16) }
}
